<?php

namespace Drupal\sdc_story_generator\Drush\Generators;

use Drupal\Core\Extension\ModuleExtensionList;
use Drupal\Core\Extension\ThemeExtensionList;
use Drupal\sdc\Plugin\Component;
use Drupal\sdc\ComponentPluginManager;
use DrupalCodeGenerator\GeneratorType;
use DrupalCodeGenerator\Attribute\Generator;
use DrupalCodeGenerator\Asset\AssetCollection;
use DrupalCodeGenerator\Command\BaseGenerator;
use Symfony\Component\Console\Question\Question;
use Symfony\Component\DependencyInjection\ContainerInterface;

#[Generator(
  name: 'theme:sdc:story',
  description: 'Generates a Storybook story from a SDC component.',
  aliases: ['story'],
  templatePath: __DIR__,
  type: GeneratorType::OTHER,
)]
final class StoryGenerator extends BaseGenerator {

  /**
   * {@inheritdoc}
   */
  protected bool $isNewExtension = FALSE;

  /**
   * {@inheritdoc}
   */
  protected string $templatePath = __DIR__;

  /**
   * The SDC Component plugin manager.
   *
   * @var \Drupal\sdc\ComponentPluginManager
   */
  protected $componentPluginManager;

  /**
   * The module extenstion list.
   *
   * @var \Drupal\Core\Extension\ModuleExtensionList
   */
  protected $moduleExtensionList;

  /**
   * The theme extension list
   *
   * @var \Drupal\Core\Extension\ThemeExtensionList
   */
  protected $themeExtensionList;

  /**
   * Constructs a StoryGenerator object.
   */
  public function __construct(ComponentPluginManager $component_plugin_manager, ModuleExtensionList $module_extension_list, ThemeExtensionList $theme_extension_list) {
    parent::__construct();
    $this->componentPluginManager = $component_plugin_manager;
    $this->moduleExtensionList = $module_extension_list;
    $this->themeExtensionList = $theme_extension_list;
  }

  public static function create(ContainerInterface $container): self {
      return new static(
        $container->get('plugin.manager.sdc'),
        $container->get('extension.list.module'),
        $container->get('extension.list.theme')
      );
  }

  /**
   * {@inheritdoc}
   */
  public function generate(&$vars, AssetCollection $assets): void {
    $vars = $this->askQuestions($vars);
    $this->generateAssets($assets);
  }

  /**
   * Asks the questions to the user.
   *
   * @param array $vars
   *   The answers array passed by reference.
   *
   * @return array
   *   The answers to the questions.
   */
  private function askQuestions(array &$vars): array {
    $component_id = $this->askComponentName();

    // We don't perform validation on askComponentName because we rely on
    // componentPluginManager->find() to throw an error if component is
    // nonexistent.
    $component = $this->componentPluginManager->find($component_id);
    $component_data = $component->metadata;

    // Path and machineName will be used on generateAssets
    // and need to be stored into $vars.
    $vars['component_path'] = $component_data->path;
    $vars['component_machine_name'] = $component_data->machineName;

    // Get the component provider (module or theme)
    $provider = $component->getPluginDefinition()['provider'];
    $vars['machine_name'] = $provider;
    $extensions = $this->getExtensions();
    $extension_name = $extensions[$provider]['name'];

    // Opinionated: the story name takes the form Extension/ComponentName.
    $vars['story_name'] = "$extension_name/$component_data->name";

    // Merge props and slots as this is a SDC concept: everything is an argType
    // for Storybook. Identify slots for later use.
    if (isset($component_data->schema)) {
      $schema = $component_data->schema;
      $props = $schema['properties'];
      $required = isset($schema['required']) ? $schema['required'] : NULL;
    } else {
      $props = [];
      $required = NULL;
    }
    $slots = array_map(function (array $slot) {
      return array_merge($slot, ['slot' => TRUE]);
    }, $component_data->slots);
    $args = array_merge($props, $slots);

    if ($args) {
      $vars['story_props'] = [];

      foreach ($args as $arg_name => $arg_data) {
        $is_slot = !empty($arg_data['slot']) && $arg_data['slot'];
        $arg_definition = [
          'name' => $arg_name,
        ];

        if (!empty($arg_data['title'])) {
          $arg_definition['title'] = $arg_data['title'];
        }

        $type = is_array($arg_data['type']) ? $arg_data['type'][0] : $arg_data['type'];
        $arg_definition['type'] = $type;

        if ($required && in_array($arg_name, $required)) {
          $arg_definition['required'] = $required;
        };

        if (!empty($arg_data['description'])) {
          $arg_definition['description'] = $arg_data['description'];
        };

        $arg_definition['category'] = $is_slot ? 'Slots' : 'Props';

        if (!empty($arg_data['enum'])) {
          $arg_definition['options'] = $arg_data['enum'];
        };

        if (!empty($arg_data['default'])) {
          $arg_definition['default'] = $arg_data['default'];
        };

        $vars['story_args'][$arg_name] = $arg_definition;
      }
    }

    return $vars;
  }

  /**
   * Gets all the modules, installed or not.
   *
   * @return Drupal\Core\Extension\Extension[]
   */
  private function getModules(): array {
    return $this->moduleExtensionList->getAllAvailableInfo();
  }

  /**
   * Gets all the themes, installed or not.
   *
   * @return Drupal\Core\Extension\Extension[]
   */
  private function getThemes() {
    return $this->themeExtensionList->getAllAvailableInfo();
  }

  private function getExtensions(): array {
    return [
      ...$this->getModules(),
      ...$this->getThemes(),
    ];
  }

  /**
   * Create the assets that the framework will write to disk later on.
   *
   * @param array $vars
   *   The answers to the CLI questions.
   */
  private function generateAssets(AssetCollection $assets): void {
    $component_path_token = '{component_path}' . DIRECTORY_SEPARATOR;
    $assets->addFile($component_path_token . "{component_machine_name}.stories.yml", 'stories-yml--template.twig');
  }

  /**
   * Asks the component name, with autocomplete helper.
   *
   * @return string
   *   The component ID.
   */
  protected function askComponentName(): string {
    $component_question = new Question('What is the component ID? (i.e: mytheme:mycomponent)', '');
    $component_ids = array_map(function (Component $component) {
      return $component->getPluginId();
    }, $this->componentPluginManager->getAllComponents());
    $component_question->setAutocompleterValues(\array_values($component_ids));
    return $this->io()->askQuestion($component_question);
  }

  /**
   * {@inheritdoc}
   */
  protected function getDestination(array $vars): string {
    // The component already informs us of the destinatoin path relative to
    // the root: that's why we bypass the original getDestination and simply
    // output the Drupal root.
    return DRUPAL_ROOT;
  }

}
